from typing import TypeAlias

from rainbowadn.collection.pair import *
from rainbowadn.core import *
from rainbowadn.flow.core import *
from rainbowadn.flow.verification.core import *
from rainbowadn.nullability import *
from ._bankflow import *
from ._flowbank import *
from ._flowblock import *
from ._flowcheque import *
from ._flowstandard import *

__all__ = ('BankBlock',)

Index: TypeAlias = FlowStandard[FlowBlock[Pair[FlowCheque, FlowBank]]]


class BankBlock:
    def __init__(self, reference: NullableReference[FlowBlock[Pair[FlowCheque, FlowBank]]]):
        assert isinstance(reference, NullableReference)
        self.reference = reference

    @classmethod
    def flow(cls) -> 'BankFlow':
        return BankFlow(FlowBank.empty())

    @classmethod
    def link_factory(cls) -> RainbowFactory[Pair[FlowCheque, FlowBank]]:
        return PairFactory(FlowCheque.factory(), FlowBank.factory()).loose()

    @classmethod
    def empty(cls) -> 'BankBlock':
        return cls(
            NullableReference(
                Null(),
                FlowBlockFactory(cls.link_factory()).loose()
            )
        )

    @classmethod
    def verification(cls) -> Verification[Index]:
        return FlowBlockVerification(cls.flow().link_verification()).loose()

    async def verify(self) -> bool:
        assert_true(await self.verification().verify(await FlowBlock.outer_of(self.link_factory(), self.reference)))
        return True

    async def add(self, cheque: FlowCheque) -> 'BankBlock':
        assert isinstance(cheque, FlowCheque)
        return await AddCheque(cheque).map(self)


class AddCheque(
    Mapper[BankBlock, BankBlock]
):
    def __init__(self, cheque: FlowCheque):
        assert isinstance(cheque, FlowCheque)
        self.cheque = cheque

    @classmethod
    async def _bank_for_link(cls, link: Nullable[Pair[FlowCheque, FlowBank]]) -> Nullable[FlowBank]:
        assert isinstance(link, Nullable)
        return Null() if link.null() else NotNull(await (link.resolve()).element1.resolve())

    async def _next_bank_for_bank(self, bank: Nullable[FlowBank]) -> FlowBank:
        return await BankBlock.flow().add(bank, self.cheque)

    async def _link_for_bank(self, bank: FlowBank) -> HashPoint[Pair[FlowCheque, FlowBank]]:
        assert isinstance(bank, FlowBank)
        return HashPoint.of(Pair(HashPoint.of(self.cheque), HashPoint.of(bank)))

    async def _next_link_for_link(
            self, link: Nullable[Pair[FlowCheque, FlowBank]]
    ) -> HashPoint[Pair[FlowCheque, FlowBank]]:
        assert isinstance(link, Nullable)
        return await self._link_for_bank(
            await self._next_bank_for_bank(
                await self._bank_for_link(link)
            )
        )

    @classmethod
    async def _add_link_to_reference(
            cls,
            link: HashPoint[Pair[FlowCheque, FlowBank]],
            reference: NullableReference[FlowBlock[Pair[FlowCheque, FlowBank]]],
    ) -> BankBlock:
        assert isinstance(link, HashPoint)
        assert isinstance(reference, NullableReference)
        return BankBlock(NullableReference.off(await FlowBlock.add_to(reference, link)))

    async def _add_to_link(
            self,
            previous: Nullable[Pair[FlowCheque, FlowBank]],
            reference: NullableReference[FlowBlock[Pair[FlowCheque, FlowBank]]],
    ) -> BankBlock:
        assert isinstance(previous, Nullable)
        assert isinstance(reference, NullableReference)
        return await self._add_link_to_reference(await self._next_link_for_link(previous), reference)

    async def _add_to_reference(
            self,
            reference: NullableReference[FlowBlock[Pair[FlowCheque, FlowBank]]],
    ) -> BankBlock:
        assert isinstance(reference, NullableReference)
        return await self._add_to_link(await FlowBlock.link_of(reference), reference)

    async def map(self, element: BankBlock) -> BankBlock:
        return await self._add_to_reference(element.reference)
